// Copyright 2016 Proyectos y Sistemas de Mantenimiento SL (eProsima).
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/*************************************************************************
 * @file Data1mb.h
 * This header file contains the declaration of the described types in the IDL file.
 *
 * This file was generated by the tool gen.
 */

#ifndef _Data1mb_H_
#define _Data1mb_H_

// TODO Poner en el contexto.

#include <stdint.h>
#include <array>
#include <string>
#include <vector>

#if defined(_WIN32)
#if defined(EPROSIMA_USER_DLL_EXPORT)
#define eProsima_user_DllExport __declspec( dllexport )
#else
#define eProsima_user_DllExport
#endif
#else
#define eProsima_user_DllExport
#endif

#if defined(_WIN32)
#if defined(EPROSIMA_USER_DLL_EXPORT)
#if defined(Data1mb_SOURCE)
#define Data1mb_DllAPI __declspec( dllexport )
#else
#define Data1mb_DllAPI __declspec( dllimport )
#endif // Data1mb_SOURCE
#else
#define Data1mb_DllAPI
#endif
#else
#define Data1mb_DllAPI
#endif // _WIN32

namespace eprosima
{
    namespace fastcdr
    {
        class Cdr;
    }
}

/*!
 * @brief This class represents the structure Data1mb defined by the user in the IDL file.
 * @ingroup DATA1MB
 */
class Data1mb
{
public:

    /*!
     * @brief Default constructor.
     */
    eProsima_user_DllExport Data1mb();
    
    /*!
     * @brief Default destructor.
     */
    eProsima_user_DllExport ~Data1mb();
    
    /*!
     * @brief Copy constructor.
     * @param x Reference to the object Data1mb that will be copied.
     */
    eProsima_user_DllExport Data1mb(const Data1mb &x);
    
    /*!
     * @brief Move constructor.
     * @param x Reference to the object Data1mb that will be copied.
     */
    eProsima_user_DllExport Data1mb(Data1mb &&x);
    
    /*!
     * @brief Copy assignment.
     * @param x Reference to the object Data1mb that will be copied.
     */
    eProsima_user_DllExport Data1mb& operator=(const Data1mb &x);
    
    /*!
     * @brief Move assignment.
     * @param x Reference to the object Data1mb that will be copied.
     */
    eProsima_user_DllExport Data1mb& operator=(Data1mb &&x);

    bool operator==(const Data1mb &x) const;
    
    /*!
     * @brief This function copies the value in member data
     * @param _data New value to be copied in member data
     */
    inline eProsima_user_DllExport void data(const std::vector<uint8_t> &_data)
    {
        m_data = _data;
    }

    /*!
     * @brief This function moves the value in member data
     * @param _data New value to be moved in member data
     */
    inline eProsima_user_DllExport void data(std::vector<uint8_t> &&_data)
    {
        m_data = std::move(_data);
    }

    /*!
     * @brief This function returns a constant reference to member data
     * @return Constant reference to member data
     */
    inline eProsima_user_DllExport const std::vector<uint8_t>& data() const
    {
        return m_data;
    }

    /*!
     * @brief This function returns a reference to member data
     * @return Reference to member data
     */
    inline eProsima_user_DllExport std::vector<uint8_t>& data()
    {
        return m_data;
    }
    
    /*!
     * @brief This function returns the maximum serialized size of an object
     * depending on the buffer alignment.
     * @param current_alignment Buffer alignment.
     * @return Maximum serialized size.
     */
    eProsima_user_DllExport static size_t getMaxCdrSerializedSize(size_t current_alignment = 0);

    eProsima_user_DllExport static size_t getCdrSerializedSize(const Data1mb& data, size_t current_alignment = 0);



    /*!
     * @brief This function serializes an object using CDR serialization.
     * @param cdr CDR serialization object.
     */
    eProsima_user_DllExport void serialize(eprosima::fastcdr::Cdr &cdr) const;

    /*!
     * @brief This function deserializes an object using CDR serialization.
     * @param cdr CDR serialization object.
     */
    eProsima_user_DllExport void deserialize(eprosima::fastcdr::Cdr &cdr);



    /*!
     * @brief This function returns the maximum serialized size of the Key of an object
     * depending on the buffer alignment.
     * @param current_alignment Buffer alignment.
     * @return Maximum serialized size.
     */
    eProsima_user_DllExport static size_t getKeyMaxCdrSerializedSize(size_t current_alignment = 0);

    /*!
     * @brief This function tells you if the Key has been defined for this type
     */
    eProsima_user_DllExport static bool isKeyDefined();

    /*!
     * @brief This function serializes the key members of an object using CDR serialization.
     * @param cdr CDR serialization object.
     */
    eProsima_user_DllExport void serializeKey(eprosima::fastcdr::Cdr &cdr) const;
    
private:
    std::vector<uint8_t> m_data;
};

#endif // _Data1mb_H_
